@@ -77,67 +77,28 @@ class HTML2FPDF(HTMLParser):
         self.theader = None  # table header cells
         self.tfooter = None  # table footer cells
         self.theader_out = self.tfooter_out = False
+        self.table_row_height = 0
+        self.heading_level = None
         self.hsize = dict(h1=2, h2=1.5, h3=1.17, h4=1, h5=0.83, h6=0.67)
+        self._only_imgs_in_td = False
 
-    def width2mm(self, length):
+    def width2unit(self, length):
+        "Handle conversion of % measures into the measurement unit used"
         if length[-1] == "%":
             total = self.pdf.w - self.pdf.r_margin - self.pdf.l_margin
             if self.table["width"][-1] == "%":
                 total *= int(self.table["width"][:-1]) / 100
-            return int(length[:-1]) * total / 101
-        return int(length) / 6
+            return int(length[:-1]) * total / 100
+        return int(length)
 
     def handle_data(self, data):
         if self.td is not None:  # drawing a table?
-            if "width" not in self.td and "colspan" not in self.td:
-                # pylint: disable=raise-missing-from
-                try:
-                    l = [self.table_col_width[self.table_col_index]]
-                except IndexError:
-                    raise ValueError(
-                        f"Width not specified for table column {self.table_col_index},"
-                        " unable to continue"
-                    )
-            elif "colspan" in self.td:
-                i = self.table_col_index
-                colspan = int(self.td["colspan"])
-                l = self.table_col_width[i : i + colspan]
-            else:
-                l = [self.td.get("width", "240")]
-            w = sum(self.width2mm(length) for length in l)
-            h = int(self.td.get("height", 0)) // 4 or self.h * 1.30
-            self.table_h = h
-            border = int(self.table.get("border", 0))
-            if not self.th:
-                align = self.td.get("align", "L")[0].upper()
-                border = border and "LR"
-            else:
-                self.set_style("B", True)
-                border = border or "B"
-                align = self.td.get("align", "C")[0].upper()
-            bgcolor = hex2dec(self.td.get("bgcolor", self.tr.get("bgcolor", "")))
-            # parsing table header/footer (drawn later):
-            if self.thead is not None:
-                self.theader.append(((w, h, data, border, 0, align), bgcolor))
-            if self.tfoot is not None:
-                self.tfooter.append(((w, h, data, border, 0, align), bgcolor))
-            # check if reached end of page, add table footer and header:
-            height = h + (self.tfooter and self.tfooter[0][0][1] or 0)
-            if self.pdf.y + height > self.pdf.page_break_trigger and not self.th:
-                self.output_table_footer()
-                self.pdf.add_page(same=True)
-                self.theader_out = self.tfooter_out = False
-            if self.tfoot is None and self.thead is None:
-                if not self.theader_out:
-                    self.output_table_header()
-                self.box_shadow(w, h, bgcolor)
-                LOGGER.debug("td cell %s %s '%s' *", self.pdf.x, w, data)
-                self.pdf.cell(w, h, data, border=border, ln=0, align=align)
+            self._insert_td(data)
         elif self.table is not None:
             # ignore anything else than td inside a table
             pass
         elif self.align:
-            LOGGER.debug("cell '%s' *", data)
+            LOGGER.debug("align '%s'", data.replace("\n", "\\n"))
             self.pdf.cell(
                 0,
                 self.h,
@@ -152,11 +113,86 @@ class HTML2FPDF(HTMLParser):
             if self.href:
                 self.put_link(self.href, data)
             else:
-                LOGGER.debug("write '%s' *", data)
+                if self.heading_level:
+                    self.pdf.start_section(data, self.heading_level - 1)
+                LOGGER.debug("write '%s'", data.replace("\n", "\\n"))
                 self.pdf.write(self.h, data)
 
+    def _insert_td(self, data=""):
+        self._only_imgs_in_td = False
+        width = self._td_width()
+        height = int(self.td.get("height", 0)) // 4 or self.h * 1.30
+        if not self.table_row_height:
+            self.table_row_height = height
+        elif self.table_row_height > height:
+            height = self.table_row_height
+        border = int(self.table.get("border", 0))
+        if self.th:
+            self.set_style("B", True)
+            border = border or "B"
+            align = self.td.get("align", "C")[0].upper()
+        else:
+            align = self.td.get("align", "L")[0].upper()
+            border = border and "LR"
+        bgcolor = hex2dec(self.td.get("bgcolor", self.tr.get("bgcolor", "")))
+        # parsing table header/footer (drawn later):
+        if self.thead is not None:
+            self.theader.append(((width, height, data, border, 0, align), bgcolor))
+        if self.tfoot is not None:
+            self.tfooter.append(((width, height, data, border, 0, align), bgcolor))
+        # check if reached end of page, add table footer and header:
+        if self.tfooter:
+            height += self.tfooter[0][0][1]
+        if self.pdf.y + height > self.pdf.page_break_trigger and not self.th:
+            self.output_table_footer()
+            self.pdf.add_page(same=True)
+            self.theader_out = self.tfooter_out = False
+        if self.tfoot is None and self.thead is None:
+            if not self.theader_out:
+                self.output_table_header()
+            self.box_shadow(width, height, bgcolor)
+            # self.pdf.x may have shifted due to <img> inside <td>:
+            self.pdf.set_x(self._td_x())
+            LOGGER.debug(
+                "td cell x=%d width=%d height=%d border=%s align=%s '%s'",
+                self.pdf.x,
+                width,
+                height,
+                border,
+                align,
+                data.replace("\n", "\\n"),
+            )
+            self.pdf.cell(width, height, data, border=border, ln=0, align=align)
+
+    def _td_x(self):
+        "Return the current table cell left side horizontal position"
+        prev_cells_total_width = sum(
+            self.width2unit(width)
+            for width in self.table_col_width[: self.table_col_index]
+        )
+        return self.table_offset + prev_cells_total_width
+
+    def _td_width(self):
+        "Return the current table cell width"
+        # pylint: disable=raise-missing-from
+        if "width" in self.td:
+            column_widths = [self.td["width"]]
+        elif "colspan" in self.td:
+            i = self.table_col_index
+            colspan = int(self.td["colspan"])
+            column_widths = self.table_col_width[i : i + colspan]
+        else:
+            try:
+                column_widths = [self.table_col_width[self.table_col_index]]
+            except IndexError:
+                raise ValueError(
+                    f"Width not specified for table column {self.table_col_index},"
+                    " unable to continue"
+                )
+        return sum(self.width2unit(width) for width in column_widths)
+
     def box_shadow(self, w, h, bgcolor):
-        LOGGER.debug("box_shadow %s %s %s", w, h, bgcolor)
+        LOGGER.debug("box_shadow w=%d h=%d bgcolor=%s", w, h, bgcolor)
         if bgcolor:
             fill_color = self.pdf.fill_color
             self.pdf.set_fill_color(*bgcolor)
@@ -193,7 +229,7 @@ class HTML2FPDF(HTMLParser):
     def output_table_sep(self):
         x1 = self.pdf.x
         y1 = self.pdf.y
-        width = sum(self.width2mm(length) for length in self.table_col_width)
+        width = sum(self.width2unit(length) for length in self.table_col_width)
         self.pdf.line(x1, y1, x1 + width, y1)
 
     def handle_starttag(self, tag, attrs):
@@ -210,6 +246,7 @@ class HTML2FPDF(HTMLParser):
             if attrs:
                 self.align = attrs.get("align")
         if tag in self.hsize:
+            self.heading_level = int(tag[1:])
             k = self.hsize[tag]
             self.pdf.ln(5 * k)
             self.pdf.set_text_color(150, 0, 0)
@@ -272,48 +309,91 @@ class HTML2FPDF(HTMLParser):
             self.tfooter = []
             self.thead = None
             self.tfoot = None
-            self.table_h = 0
             self.pdf.ln()
         if tag == "tr":
             self.tr_index = 0 if self.tr_index is None else (self.tr_index + 1)
             self.tr = {k.lower(): v for k, v in attrs.items()}
             self.table_col_index = 0
+            self.table_row_height = 0
             self.pdf.set_x(self.table_offset)
             # Adding an horizontal line separator between rows:
             if self.table_line_separators and self.tr_index > 0:
                 self.output_table_sep()
         if tag == "td":
             self.td = {k.lower(): v for k, v in attrs.items()}
+            if "width" in self.td and self.table_col_index >= len(self.table_col_width):
+                assert self.table_col_index == len(
+                    self.table_col_width
+                ), f"table_col_index={self.table_col_index} #table_col_width={len(self.table_col_width)}"
+                self.table_col_width.append(self.td["width"])
+            if attrs:
+                self.align = attrs.get("align")
+            self._only_imgs_in_td = False
         if tag == "th":
             self.td = {k.lower(): v for k, v in attrs.items()}
             self.th = True
-            if "width" in self.td:
+            if "width" in self.td and self.table_col_index >= len(self.table_col_width):
+                assert self.table_col_index == len(
+                    self.table_col_width
+                ), f"table_col_index={self.table_col_index} #table_col_width={len(self.table_col_width)}"
                 self.table_col_width.append(self.td["width"])
         if tag == "thead":
             self.thead = {}
         if tag == "tfoot":
             self.tfoot = {}
         if tag == "img" and "src" in attrs:
-            w = px2mm(attrs.get("width", 0))
-            h = px2mm(attrs.get("height", 0))
-            if self.pdf.y + h > self.pdf.page_break_trigger:
+            width = px2mm(attrs.get("width", 0))
+            height = px2mm(attrs.get("height", 0))
+            if self.pdf.y + height > self.pdf.page_break_trigger:
                 self.pdf.add_page(same=True)
-            x = self.pdf.get_x()
             y = self.pdf.get_y()
-            if self.align and self.align[0].upper() == "C":
-                x = self.pdf.w / 2 - w / 2
-            self.pdf.image(self.image_map(attrs["src"]), x, y, w, h, link=self.href)
-            self.pdf.set_x(x + w)
-            self.pdf.set_y(y + h)
+            if self.table_col_index is not None:
+                self._only_imgs_in_td = True
+                # <img> in a <td>: its width must not exceed the cell width:
+                td_width = self._td_width()
+                if not width or width > td_width:
+                    if width:  # Preserving image aspect ratio:
+                        height *= td_width / width
+                    width = td_width
+                x = self._td_x()
+                if self.align and self.align[0].upper() == "C":
+                    x += (td_width - width) / 2
+            else:
+                x = self.pdf.get_x()
+                if self.align and self.align[0].upper() == "C":
+                    x = self.pdf.w / 2 - width / 2
+            LOGGER.debug(
+                'image "%s" x=%d y=%d width=%d height=%d',
+                attrs["src"],
+                x,
+                y,
+                width,
+                height,
+            )
+            self.pdf.image(
+                self.image_map(attrs["src"]), x, y, width, height, link=self.href
+            )
+            self.pdf.set_x(x + width)
+            if self.table_col_index is not None:
+                # <img> in a <td>: we grow the cell height according to the image height:
+                if height > self.table_row_height:
+                    self.table_row_height = height
+            else:
+                self.pdf.set_y(y + height)
         if tag in ("b", "i", "u"):
             self.set_style(tag, True)
         if tag == "center":
             self.align = "Center"
+        if tag == "toc":
+            self.pdf.insert_toc_placeholder(
+                self.render_toc, pages=int(attrs.get("pages", 1))
+            )
 
     def handle_endtag(self, tag):
         # Closing tag
         LOGGER.debug("ENDTAG %s", tag)
-        if tag in ("h1", "h2", "h3", "h4", "h5", "h6"):
+        if tag in self.hsize:
+            self.heading_level = None
             self.pdf.ln(self.h)
             self.set_font()
             self.set_text_color()
@@ -360,12 +440,15 @@ class HTML2FPDF(HTMLParser):
             self.tr_index = None
         if tag == "tr":
             if self.tfoot is None:
-                self.pdf.ln(self.table_h)
+                self.pdf.ln(self.table_row_height)
+            self.table_col_index = None
             self.tr = None
         if tag in ("td", "th"):
             if self.th:
                 LOGGER.debug("revert style")
                 self.set_style("b", False)  # revert style
+            elif self._only_imgs_in_td:
+                self._insert_td()
             self.table_col_index += int(self.td.get("colspan", "1"))
             self.td = None
             self.th = False
@@ -413,6 +496,17 @@ class HTML2FPDF(HTMLParser):
         self.set_style("u", False)
         self.set_text_color()
 
+    # pylint: disable=no-self-use
+    def render_toc(self, pdf, outline):
+        "This method can be overriden by subclasses to customize the Table of Contents style."
+        pdf.ln()
+        for section in outline[1:]:  # skipping top-level h1
+            link = pdf.add_link()
+            pdf.set_link(link, page=section.page_number)
+            text = f'{" " * section.level * 2} {section.name}'
+            text += f' {"." * (60 - section.level*2 - len(section.name))} {section.page_number}'
+            pdf.multi_cell(w=pdf.epw, h=pdf.font_size, txt=text, ln=1, link=link)
+
     # Subclasses of _markupbase.ParserBase must implement this:
     def error(self, message):
         raise RuntimeError(message)
